# coding=utf-8
import setuptools

setuptools.setup(name='shy-tools',
                 version='0.0.9',
                 author='shy',
                 author_email='13141462249@163.com',
                 url='https://github.com/smileshy777/shy_tools',
                 description='My code tools',
                 packages=setuptools.find_packages())
